==> newsstar-1.2.2.tar.gz: newsstar-1.2.2/src/rw.c <==
/*
    newsstar - advanced news fetcher
    Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Routines for reading and writing files etc */

/* $Id: rw.c,v 1.5 2004/07/03 21:21:24 realh Exp $ */

#include "common.h"

#include <errno.h>
#include <sys/select.h>

#include "child.h"
#include "stats.h"

#define BUFLEN 256

/* Minimum number of bytes to read into at end of buffer */
#define MINREAD 16

#ifdef SUPPORT_SSL
rw_buf *rw_create_buf_ssl(int fd, SSL *ssl, BOOL crlf,
	BOOL threadsafe, int thread_idx)
#else
rw_buf *rw_create_buf(int fd, BOOL crlf, BOOL threadsafe, int thread_idx)
#endif
{
    rw_buf *buf;
    
    buf = XMALLOC(sizeof(rw_buf));
    buf->buf = XMALLOC(buf->buflen = BUFLEN);
    buf->buf[0] = 0;
    buf->fd = fd;
    buf->surplus_start = buf->surplus_size = 0;
    buf->status = rw_Normal;
    buf->crlf = crlf;
    buf->threadsafe = threadsafe;
    buf->thread_idx = thread_idx;
    buf->rcvd = buf->sent = 0;
    buf->scrutinise = FALSE;
    if (buf->thread_idx != -1)
    {
	struct timezone tz;
	
	gettimeofday(&buf->tv_read, &tz);
	buf->tv_write = buf->tv_read;
    }
#ifdef SUPPORT_SSL
    buf->ssl = ssl;
#endif
    return buf;
}

void rw_free_buf(rw_buf *buf)
{
    if (buf)
	free(buf->buf);
    free(buf);
}

int rw_close(rw_buf *buf)
{
    int result = 0;

    if (!buf)
	return -1;
#ifdef SUPPORT_SSL
    if (buf->ssl)
    {
	if (!SSL_shutdown(buf->ssl))
	{
	    result = SSL_shutdown(buf->ssl);
	    if (result == 1)
		result = 0;
	    else if (!result)
		result = -1;
	}
	SSL_free(buf->ssl);
    }
#endif
    if (buf->fd > -1)
    {
	result = result || close(buf->fd);
	buf->fd = -1;
    }
    else
    {
	deblog_warning(_("Closing already closed rw_buf file descriptor\n"));
    }
    return result;
}

int rw_close_and_free(rw_buf *buf)
{
    int result;
    
    result = rw_close(buf);

    rw_free_buf(buf);
    return result;
}

/* Size of terminator used by a particular rw_buf */
static size_t rw_term_size(const rw_buf *buf)
{
    return buf->crlf ? 2 : 1;
}

/* If a string contains a terminator within len, returns its offset,
 * otherwise -1 
 */
static int rw_str_term(const rw_buf *buf, const char *str, size_t len)
{
    size_t i;
    for (i = 0; i < len; ++i)
    {
	if (str[i] == '\n')
        {
            if (buf->crlf)
            {
                if (str[i - 1] == '\r')
                    return i - 1;
            }
            else
	        return i;
        }
    }
    return -1;
}

/* TRANSFER_STATS_INTERVAL is how often (in ms) stats on bytes transferred per
 * thread should be sent to parent */
#ifdef TRANSFER_STATS_INTERVAL
#define INTERVAL_TV \
 { (TRANSFER_STATS_INTERVAL) / 1000, ((TRANSFER_STATS_INTERVAL) * 1000) % 1000 }
#else
#define INTERVAL_TV { 0, 200000 }
#endif
static struct timeval rw_interval = INTERVAL_TV;

static void rw_send_stats(intercom_Command cmd, struct timeval *last_time, long
	int *amount)
{
    struct timeval now;
    struct timeval difference;
    struct timezone tz;

    gettimeofday(&now, &tz);
    /* deblog_debug(_("rw_send_stats %ld at %.2f\n"), *amount, FLOAT_TV(&now));
     */
    /* Subtract last_time from now */
    xtimeval_subtract(&now, last_time, &difference);
    if (timercmp(&difference, &rw_interval, > ))
    {
	child_send_command(cmd, "%ld %ld %ld", *amount, last_time->tv_sec,
		last_time->tv_usec);
	*last_time = now;
	*amount = 0;
    }
}

const char *rw_read_line(rw_buf *buf)
{
    static FILE *rrl_log = NULL;

    if (buf->scrutinise && !rrl_log)
    {
	rrl_log = fopen("/var/log/news/rrl", "w");
	if (!rrl_log)
	    buf->scrutinise = FALSE;
    }
    if (buf->status)
	return NULL;
    if (buf->surplus_size)
    {
	if (buf->scrutinise)
	{
	    fprintf(rrl_log, "Shifting %d surplus from %d to start\n",
		    buf->surplus_size, buf->surplus_start);
	}
	memmove(buf->buf, buf->buf + buf->surplus_start, buf->surplus_size);
	buf->surplus_start = 0;
	if (buf->scrutinise)
	{
	    fprintf(rrl_log, "%.*s\n", (int) buf->surplus_size, buf->buf);
	    fprintf(rrl_log,
		    (rw_str_term(buf, buf->buf, buf->surplus_size) == -1)
		    ? "unterminated\n" : "terminated\n");
	}
    }
    else if (buf->scrutinise)
    {
	fprintf(rrl_log, "No surplus\n");
    }
    for (;;)
    {
        int n = -1;
        int toff = rw_str_term(buf, buf->buf, buf->surplus_size);
	int read_errno = 0;
	BOOL blocked = FALSE;

        if (toff != -1)
        {
	    buf->buf[toff] = 0;
	    if (buf->scrutinise)
	    {
		fprintf(rrl_log,
			"%d terminated string left over from last read\n",
			buf->surplus_size);
		fprintf(rrl_log, "%s\n", buf->buf);
	    }
    	    buf->surplus_start = toff + rw_term_size(buf);
	    buf->surplus_size -= toff + rw_term_size(buf);
	    if (buf->scrutinise)
	    {
		fprintf(rrl_log, "Leaving %d surplus for next time\n",
			buf->surplus_size);
		if (buf->surplus_size)
		{
		    fprintf(rrl_log, "%.*s\n",
			    (int) buf->surplus_size,
			    buf->buf + buf->surplus_start);
		    fprintf(rrl_log,
			    (rw_str_term(buf, buf->buf + buf->surplus_start,
					 buf->surplus_size) == -1)
			    ? "unterminated\n" : "terminated\n");
		}
	    }
	    return buf->buf;
        }
	else if (buf->scrutinise)
	{
	    fprintf(rrl_log, "%d surplus unterminated, reading more\n",
		    buf->surplus_size);
	    fprintf(rrl_log, "%.*s\n", (int) buf->surplus_size, buf->buf);
	}
        if (buf->buflen - buf->surplus_size < MINREAD)
        {
	    if (buf->scrutinise)
	    {
		fprintf(rrl_log,
		    "Only %d left in free buffer, expanding from %d by %d\n",
		    buf->buflen - buf->surplus_size, buf->surplus_size, BUFLEN);
	    }
	    buf->buf = XREALLOC(buf->buf, buf->buflen += BUFLEN);
        }
	do
	{
	#ifdef SUPPORT_SSL
	    int ssl_err = 0;
	#endif

	    /* Rather dodgy double compile- and run-time conditional here */
	    blocked = FALSE;
	#ifdef SUPPORT_SSL
	    if (buf->ssl)
	    {
		n = SSL_read(buf->ssl, buf->buf + buf->surplus_size, 
			XMIN(buf->buflen - buf->surplus_size, BUFLEN));
		if (n < 0)
		{
		    ssl_err = SSL_get_error(buf->ssl, n);
		    read_errno = errno;
		    blocked = ssl_err == SSL_ERROR_WANT_READ ||
			ssl_err == SSL_ERROR_WANT_WRITE;
		}
	    }
	    else
	#endif
	    {
		errno = 0;
		n = read(buf->fd, buf->buf + buf->surplus_size, 
			XMIN(buf->buflen - buf->surplus_size, BUFLEN));
		read_errno = errno;
		blocked = read_errno == EWOULDBLOCK || read_errno == EAGAIN;
	    }
	    if (blocked)
	    {
		fd_set our_fds, excep_fds;
		fd_set *read_fds, *write_fds;
		struct timeval timeout;

	    #ifdef SUPPORT_SSL
		if (buf->ssl && ssl_err == SSL_ERROR_WANT_WRITE)
		{
		    write_fds = &our_fds;
		    read_fds = NULL;
		}
		else
	    #endif
		{
		    write_fds = NULL;
		    read_fds = &our_fds;
		}
		/*deblog_debug(_("Blocking on read: %d\n"), ++blocked_count);*/
		FD_ZERO(&our_fds);
		FD_SET(buf->fd, &our_fds);
		FD_ZERO(&excep_fds);
		FD_SET(buf->fd, &excep_fds);
		timeout = stats_interval;
		select(buf->fd + 1, read_fds, write_fds, &excep_fds, &timeout);
		if (buf->thread_idx == -1)
		    stats_update(FALSE);
	    }
	} while (n == -1 && blocked);
	if (!n)
	{
	    size_t old_surplus = buf->surplus_size;

	    buf->status = rw_Eof;
	    buf->buf[old_surplus] = 0;
	    buf->surplus_start = buf->surplus_size = 0;
	    if (buf->scrutinise)
	    {
		fprintf(rrl_log, "Read nothing, must be EOF\n");
		if (old_surplus)
		    fprintf(rrl_log, "Forced termination of:\n%s\n", buf->buf);
	    }
	    return old_surplus ? buf->buf : NULL;
	}
	if (n < 0)
	{
	    deblog_info(_("read error: %s (%d)\n"), strerror(errno), errno);
	    buf->status = rw_Err;
	    return NULL;
	}
	if (buf->scrutinise)
	{
	    fprintf(rrl_log, "Added another %d to buffer\n%.*s\n",
		    n, n, buf->buf + buf->surplus_size);
	    fprintf(rrl_log,
		    (rw_str_term(buf, buf->buf + buf->surplus_size, n) == -1)
		    ? "unterminated\n" : "terminated\n");
	}
	buf->surplus_size += n;
	if (buf->scrutinise)
	{
	    fprintf(rrl_log, "Entire buffer now contains %d\n%.*s\n",
		    (int) buf->surplus_size, (int) buf->surplus_size, buf->buf);
	}
	if (buf->thread_idx != -1)
	{
	    buf->rcvd += n;
	    rw_send_stats(command_AmountRcvd, &buf->tv_read, &buf->rcvd);
	}
    }
    return NULL;
}

int rw_buf_eof(const rw_buf *buf)
{
    return (buf->status == rw_Eof);
}

int rw_write_line(rw_buf *buf, const char *str)
{
    int result;
    char *termstr = XSTRDUPCAT(str, buf->crlf ? "\r\n" : "\n");

    result = rw_write(buf, termstr);
    free(termstr);
    return result;
}

int rw_write(rw_buf *buf, const char *str)
{
    const char *s = str;
    unsigned int l = strlen(str);
    int n = 0;

    do 
    {
	errno = 0;
    #ifdef SUPPORT_SSL
	if (buf->ssl)
	{
	    n = SSL_write(buf->ssl, s, l);
	}
	else
    #endif
	{
	    n = write(buf->fd, s, l);
	}
	if (n < 0)
	{
	    return n;
	}
	l -= n;
	s += n;
	if (buf->thread_idx != -1)
	{
	    buf->sent += n;
	}
	if (l > 0 && buf->threadsafe)
	{
	    /* Write this directly to stderr, because we'll almost
	     * certainly be in a subthread and can no longer trust deblog
	     */
	    fprintf(stderr,
		    _("\nError: Unable to write complete line to pipe:\n"
			"%s\n"),
		    strerror(errno));
	    exit(1);
	}
    } while (l > 0);
    if (buf->thread_idx != -1)
    {
	rw_send_stats(command_AmountSent, &buf->tv_write, &buf->sent);
    }
    return n;
}

int rw_ftos(rw_buf *src, rw_buf *dest, rw_modifier modifier, void *handle)
{
    const char *ln;

    for (ln = rw_read_line(src); ln; ln = rw_read_line(src))
    {
        if (modifier)
            ln = modifier(ln, handle);
        if (ln && ln[0] == '.')
        {
            int n;

            do 
            {
                char esc = '.';

                n = write(dest->fd, &esc, 1);
                if (n < 0)
                    return n;
            } while (!n);
        }
        if (ln && rw_write_line(dest, ln) < 0)
            return -1;
    }
    if (!rw_buf_eof(src))
        return -1;
    /* Should write 3 bytes including EOL */
    return rw_write_line(dest, ".") == 3 ? 0 : -1;
}

int rw_stof(rw_buf *src, rw_buf *dest, rw_modifier modifier, void *handle)
{
    const char *ln;

    for (ln = rw_read_line(src); ln; ln = rw_read_line(src))
    {
        const char *escln = ln;

        if (ln[0] == '.')
        {
            if (!ln[1])
                return 0;
            ++escln;
        }
        if (modifier)
            escln = modifier(escln, handle);
        if (escln && rw_write_line(dest, escln) < 0)
            return -1;
    }
    /* Shouldn't get EOF */
    return -1;
}

int rw_copy(rw_buf *src, rw_buf *dest, rw_modifier modifier, void *handle)
{
    const char *ln;

    for (ln = rw_read_line(src); ln; ln = rw_read_line(src))
    {
        const char *escln = ln;

        if (modifier)
            escln = modifier(escln, handle);
        if (escln && rw_write_line(dest, escln) < 0)
            return -1;
    }
    /* Assume EOF */
    return 0;
}

int rw_writef(rw_buf *buf, const char *format, ...)
{
    va_list ap;
    int result;

    va_start(ap, format);
    result = rw_vwritef(buf, format, ap);
    va_end(ap);
    return result;
}

int rw_vwritef(rw_buf *buf, const char *format, va_list ap)
{
    char *formatted;
    int result;

    if (vasprintf(&formatted, format, ap) < 0)
    {
	deblog_error(_("Unable to format string for sending\n"));
    }
    result = rw_write(buf, formatted);
    free(formatted);
    return result;
}

==> newsstar-1.3.2.tar.gz: newsstar-1.3.2/src/rw.c <==
/*
    newsstar - advanced news fetcher
    Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Routines for reading and writing files etc */


#include "common.h"

#include <errno.h>
#include <sys/select.h>

#include "child.h"
#include "stats.h"

#define BUFLEN 256l

/* Minimum number of bytes to read into at end of buffer */
#define MINREAD 16

#ifdef SUPPORT_SSL
rw_buf *rw_create_buf_ssl(int fd, SSL *ssl, BOOL crlf,
	BOOL threadsafe, int thread_idx)
#else
rw_buf *rw_create_buf(int fd, BOOL crlf, BOOL threadsafe, int thread_idx)
#endif
{
    rw_buf *buf;
    
    buf = XMALLOC(sizeof(rw_buf));
    buf->buf = XMALLOC(buf->buflen = BUFLEN);
    buf->buf[0] = 0;
    buf->fd = fd;
    buf->surplus_start = buf->surplus_size = 0;
    buf->status = rw_Normal;
    buf->crlf = crlf;
    buf->threadsafe = threadsafe;
    buf->thread_idx = thread_idx;
    buf->rcvd = buf->sent = 0;
    buf->scrutinise = FALSE;
    if (buf->thread_idx != -1)
    {
	struct timezone tz;
	
	gettimeofday(&buf->tv_read, &tz);
	buf->tv_write = buf->tv_read;
    }
#ifdef SUPPORT_SSL
    buf->ssl = ssl;
#endif
    return buf;
}

void rw_free_buf(rw_buf *buf)
{
    if (buf)
	free(buf->buf);
    free(buf);
}

int rw_close(rw_buf *buf)
{
    int result = 0;

    if (!buf)
	return -1;
#ifdef SUPPORT_SSL
    if (buf->ssl)
    {
	if (!SSL_shutdown(buf->ssl))
	{
	    result = SSL_shutdown(buf->ssl);
	    if (result == 1)
		result = 0;
	    else if (!result)
		result = -1;
	}
	SSL_free(buf->ssl);
    }
#endif
    if (buf->fd > -1)
    {
	result = result || close(buf->fd);
	buf->fd = -1;
    }
    else
    {
	deblog_warning(_("Closing already closed rw_buf file descriptor\n"));
    }
    return result;
}

int rw_close_and_free(rw_buf *buf)
{
    int result;
    
    result = rw_close(buf);

    rw_free_buf(buf);
    return result;
}

/* Size of terminator used by a particular rw_buf */
static size_t rw_term_size(const rw_buf *buf)
{
    return buf->crlf ? 2 : 1;
}

/* If a string contains a terminator within len, returns its offset,
 * otherwise -1 
 */
static int rw_str_term(const rw_buf *buf, const char *str, size_t len)
{
    size_t i;
    for (i = 0; i < len; ++i)
    {
	if (str[i] == '\n')
        {
            if (buf->crlf)
            {
                if (str[i - 1] == '\r')
                    return i - 1;
            }
            else
	        return i;
        }
    }
    return -1;
}

/* TRANSFER_STATS_INTERVAL is how often (in ms) stats on bytes transferred per
 * thread should be sent to parent */
#ifdef TRANSFER_STATS_INTERVAL
#define INTERVAL_TV \
 { (TRANSFER_STATS_INTERVAL) / 1000, ((TRANSFER_STATS_INTERVAL) * 1000) % 1000 }
#else
#define INTERVAL_TV { 0, 200000 }
#endif
static struct timeval rw_interval = INTERVAL_TV;

static void rw_send_stats(intercom_Command cmd, struct timeval *last_time, long
	int *amount)
{
    struct timeval now;
    struct timeval difference;
    struct timezone tz;

    gettimeofday(&now, &tz);
    /* deblog_debug(_("rw_send_stats %ld at %.2f\n"), *amount, FLOAT_TV(&now));
     */
    /* Subtract last_time from now */
    xtimeval_subtract(&now, last_time, &difference);
    if (timercmp(&difference, &rw_interval, > ))
    {
	child_send_command(cmd, "%ld %ld %ld", *amount, last_time->tv_sec,
		last_time->tv_usec);
	*last_time = now;
	*amount = 0;
    }
}

const char *rw_read_line(rw_buf *buf)
{
    static FILE *rrl_log = NULL;

    if (buf->scrutinise && !rrl_log)
    {
	rrl_log = fopen("/var/log/news/rrl", "w");
	if (!rrl_log)
	    buf->scrutinise = FALSE;
    }
    if (buf->status)
	return NULL;
    if (buf->surplus_size)
    {
	if (buf->scrutinise)
	{
	    fprintf(rrl_log, "Shifting %ld surplus from %ld to start\n",
		    buf->surplus_size, buf->surplus_start);
	}
	memmove(buf->buf, buf->buf + buf->surplus_start, buf->surplus_size);
	buf->surplus_start = 0;
	if (buf->scrutinise)
	{
	    fprintf(rrl_log, "%.*s\n", (int) buf->surplus_size, buf->buf);
	    fprintf(rrl_log,
		    (rw_str_term(buf, buf->buf, buf->surplus_size) == -1)
		    ? "unterminated\n" : "terminated\n");
	}
    }
    else if (buf->scrutinise)
    {
	fprintf(rrl_log, "No surplus\n");
    }
    for (;;)
    {
        int n = -1;
        int toff = rw_str_term(buf, buf->buf, buf->surplus_size);
	int read_errno = 0;
	BOOL blocked = FALSE;

        if (toff != -1)
        {
	    buf->buf[toff] = 0;
	    if (buf->scrutinise)
	    {
		fprintf(rrl_log,
			"%ld terminated string left over from last read\n",
			buf->surplus_size);
		fprintf(rrl_log, "%s\n", buf->buf);
	    }
    	    buf->surplus_start = toff + rw_term_size(buf);
	    buf->surplus_size -= toff + rw_term_size(buf);
	    if (buf->scrutinise)
	    {
		fprintf(rrl_log, "Leaving %ld surplus for next time\n",
			buf->surplus_size);
		if (buf->surplus_size)
		{
		    fprintf(rrl_log, "%.*s\n",
			    (int) buf->surplus_size,
			    buf->buf + buf->surplus_start);
		    fprintf(rrl_log,
			    (rw_str_term(buf, buf->buf + buf->surplus_start,
					 buf->surplus_size) == -1)
			    ? "unterminated\n" : "terminated\n");
		}
	    }
	    return buf->buf;
        }
	else if (buf->scrutinise)
	{
	    fprintf(rrl_log, "%ld surplus unterminated, reading more\n",
		    buf->surplus_size);
	    fprintf(rrl_log, "%.*s\n", (int) buf->surplus_size, buf->buf);
	}
        if (buf->buflen - buf->surplus_size < MINREAD)
        {
	    if (buf->scrutinise)
	    {
		fprintf(rrl_log,
		    "Only %ld left in free buffer, expanding from %ld by %ld\n",
		    buf->buflen - buf->surplus_size, buf->surplus_size, BUFLEN);
	    }
	    buf->buf = XREALLOC(buf->buf, buf->buflen += BUFLEN);
        }
	do
	{
	#ifdef SUPPORT_SSL
	    int ssl_err = 0;
	#endif

	    /* Rather dodgy double compile- and run-time conditional here */
	    blocked = FALSE;
	#ifdef SUPPORT_SSL
	    if (buf->ssl)
	    {
		n = SSL_read(buf->ssl, buf->buf + buf->surplus_size, 
			XMIN(buf->buflen - buf->surplus_size, BUFLEN));
		if (n < 0)
		{
		    ssl_err = SSL_get_error(buf->ssl, n);
		    read_errno = errno;
		    blocked = ssl_err == SSL_ERROR_WANT_READ ||
			ssl_err == SSL_ERROR_WANT_WRITE;
		}
	    }
	    else
	#endif
	    {
		errno = 0;
		n = read(buf->fd, buf->buf + buf->surplus_size, 
			XMIN(buf->buflen - buf->surplus_size, BUFLEN));
		read_errno = errno;
		blocked = read_errno == EWOULDBLOCK || read_errno == EAGAIN;
	    }
	    if (blocked)
	    {
		fd_set our_fds, excep_fds;
		fd_set *read_fds, *write_fds;
		struct timeval timeout;

	    #ifdef SUPPORT_SSL
		if (buf->ssl && ssl_err == SSL_ERROR_WANT_WRITE)
		{
		    write_fds = &our_fds;
		    read_fds = NULL;
		}
		else
	    #endif
		{
		    write_fds = NULL;
		    read_fds = &our_fds;
		}
		/*deblog_debug(_("Blocking on read: %d\n"), ++blocked_count);*/
		FD_ZERO(&our_fds);
		FD_SET(buf->fd, &our_fds);
		FD_ZERO(&excep_fds);
		FD_SET(buf->fd, &excep_fds);
		timeout = stats_interval;
		select(buf->fd + 1, read_fds, write_fds, &excep_fds, &timeout);
		if (buf->thread_idx == -1)
		    stats_update(FALSE);
	    }
	} while (n == -1 && blocked);
	if (!n)
	{
	    size_t old_surplus = buf->surplus_size;

	    buf->status = rw_Eof;
	    buf->buf[old_surplus] = 0;
	    buf->surplus_start = buf->surplus_size = 0;
	    if (buf->scrutinise)
	    {
		fprintf(rrl_log, "Read nothing, must be EOF\n");
		if (old_surplus)
		    fprintf(rrl_log, "Forced termination of:\n%s\n", buf->buf);
	    }
	    return old_surplus ? buf->buf : NULL;
	}
	if (n < 0)
	{
	    deblog_info(_("read error: %s\n"), strerror(read_errno));
	    buf->status = rw_Err;
	    return NULL;
	}
	if (buf->scrutinise)
	{
	    fprintf(rrl_log, "Added another %d to buffer\n%.*s\n",
		    n, n, buf->buf + buf->surplus_size);
	    fprintf(rrl_log,
		    (rw_str_term(buf, buf->buf + buf->surplus_size, n) == -1)
		    ? "unterminated\n" : "terminated\n");
	}
	buf->surplus_size += n;
	if (buf->scrutinise)
	{
	    fprintf(rrl_log, "Entire buffer now contains %d\n%.*s\n",
		    (int) buf->surplus_size, (int) buf->surplus_size, buf->buf);
	}
	if (buf->thread_idx != -1)
	{
	    buf->rcvd += n;
	    rw_send_stats(command_AmountRcvd, &buf->tv_read, &buf->rcvd);
	}
    }
    return NULL;
}

int rw_buf_eof(const rw_buf *buf)
{
    return (buf->status == rw_Eof);
}

int rw_write_line(rw_buf *buf, const char *str)
{
    int result;
    char *termstr = XSTRDUPCAT(str, buf->crlf ? "\r\n" : "\n");

    result = rw_write(buf, termstr);
    free(termstr);
    return result;
}

int rw_write(rw_buf *buf, const char *str)
{
    const char *s = str;
    unsigned int l = strlen(str);
    int n = 0;

    do 
    {
	errno = 0;
    #ifdef SUPPORT_SSL
	if (buf->ssl)
	{
	    n = SSL_write(buf->ssl, s, l);
	}
	else
    #endif
	{
	    n = write(buf->fd, s, l);
	}
	if (n < 0)
	{
	    return n;
	}
	l -= n;
	s += n;
	if (buf->thread_idx != -1)
	{
	    buf->sent += n;
	}
	if (l > 0 && buf->threadsafe)
	{
	    /* Write this directly to stderr, because we'll almost
	     * certainly be in a subthread and can no longer trust deblog
	     */
	    fprintf(stderr,
		    _("\nError: Unable to write complete line to pipe:\n"
			"%s\n"),
		    strerror(errno));
	    exit(1);
	}
    } while (l > 0);
    if (buf->thread_idx != -1)
    {
	rw_send_stats(command_AmountSent, &buf->tv_write, &buf->sent);
    }
    return n;
}

int rw_ftos(rw_buf *src, rw_buf *dest, rw_modifier modifier, void *handle)
{
    const char *ln;

    for (ln = rw_read_line(src); ln; ln = rw_read_line(src))
    {
        if (modifier)
            ln = modifier(ln, handle);
        if (ln && ln[0] == '.')
        {
            int n;

            do 
            {
                char esc = '.';

                n = write(dest->fd, &esc, 1);
                if (n < 0)
                    return n;
            } while (!n);
        }
        if (ln && rw_write_line(dest, ln) < 0)
            return -1;
    }
    if (!rw_buf_eof(src))
        return -1;
    /* Should write 3 bytes including EOL */
    return rw_write_line(dest, ".") == 3 ? 0 : -1;
}

int rw_stof(rw_buf *src, rw_buf *dest, rw_modifier modifier, void *handle)
{
    const char *ln;

    for (ln = rw_read_line(src); ln; ln = rw_read_line(src))
    {
        const char *escln = ln;

        if (ln[0] == '.')
        {
            if (!ln[1])
                return 0;
            ++escln;
        }
        if (modifier)
            escln = modifier(escln, handle);
        if (escln && rw_write_line(dest, escln) < 0)
            return -1;
    }
    /* Shouldn't get EOF */
    return -1;
}

int rw_copy(rw_buf *src, rw_buf *dest, rw_modifier modifier, void *handle)
{
    const char *ln;

    for (ln = rw_read_line(src); ln; ln = rw_read_line(src))
    {
        const char *escln = ln;

        if (modifier)
            escln = modifier(escln, handle);
        if (escln && rw_write_line(dest, escln) < 0)
            return -1;
    }
    /* Assume EOF */
    return 0;
}

int rw_writef(rw_buf *buf, const char *format, ...)
{
    va_list ap;
    int result;

    va_start(ap, format);
    result = rw_vwritef(buf, format, ap);
    va_end(ap);
    return result;
}

int rw_vwritef(rw_buf *buf, const char *format, va_list ap)
{
    char *formatted;
    int result;

    if (vasprintf(&formatted, format, ap) < 0)
    {
	deblog_error(_("Unable to format string for sending\n"));
    }
    result = rw_write(buf, formatted);
    free(formatted);
    return result;
}
